import java.io.*;
import java.text.*;
import java.util.*;

/**
 * Pascal-2001 Optimizing Compiler with Type-Checking<BR>
 * ICS 142/ICS 145A<BR>
 * Justin Erenkrantz<BR>
 * 12/3/2000<BR>
 * 
 * <P>This class will compile a Pascal-2001 file.
 *
 * <P>This compiler will do a variety of optimizations 
 * (including register optimizations, and constant folding).
 *
 * <P>The source file name is read in from the command line.
 *
 * <P>Generated instruction mneumonics go to .S.
 *
 * <P>Generated instructions go to .o.
 *
 */
public class PascalParser
{
   /**
    * Base Exception class.
    */
   private class ParsingException extends Exception
   {
   }

   /**
    * This is an Exception that is thrown
    * if a type error occurs while parsing
    */
   private class TypeException extends ParsingException
   {
      private PascalSymbol leftSymbol;
      private PascalSymbol rightSymbol;
      private PascalType rightType;
      private String undeclaredName;

      public TypeException(PascalSymbol newLeft, PascalSymbol newRight)
      {
         create(newLeft, newRight, null, null);
      }   

      public TypeException(PascalSymbol newLeft, PascalType newRight)
      {
         create(newLeft, null, newRight, null);
      }   

      public TypeException(PascalSymbol newLeft)
      {
         create(newLeft, null, null, null);
      }

      public TypeException(String s)
      {
         create(null, null, null, s);
      }   

      private void create(PascalSymbol ls, PascalSymbol rs, PascalType rt, String un)
      {
         leftSymbol = ls;
         rightSymbol = rs;   
         rightType = rt;
         undeclaredName = un;
      
      }

      public String toString()
      {
         if (undeclaredName != null)
            return undeclaredName + " not found";
         else if (rightType != null)
            return "Incorrect type (" + leftSymbol.type.form + " / " +  rightType.form + ")\nLeft Symbol:\n" + leftSymbol + "\n\nRight Type:\n" + rightType;
         else if (rightSymbol != null)
            return "Left:\n" + leftSymbol + "\nRight:\n" + rightSymbol + " do not match";
         else 
            return leftSymbol.identifier + " is not used in valid expression\n" + leftSymbol;
      }
   }

   /**
    * This is a Syntax Exception that is thrown
    * if some error occurs while parsing
    */
   private class SyntaxException extends ParsingException
   {
      private int badToken;
      private int wantedToken;
      private String badTokenString;
      private String wantedTokenString;

      public SyntaxException(int myBadToken, int myWantedToken)
      {
         badToken = myBadToken;
         wantedToken = myWantedToken;
         badTokenString = PascalLanguage.keywords[badToken];
         wantedTokenString = PascalLanguage.keywords[wantedToken];
      }   

      public String toString()
      {
         return "Found: " + badToken + " ("+ badTokenString + ") Wanted: " 
                   + wantedToken + " (" + wantedTokenString + ")";
      }
   }

   /**
    * Our scanner instance
    */   
   private PascalScanner ps;

   /**
    * Our target architecture instance
    */
   private TargetArchitecture ta;

   /**
    * Number of type-based errors seen so far in the current source file.
    */
   private int typeErrorCount;

   /**
    * If set, allow to backtrack one token.
    * Allows LL(1) grammars.
    */
   private boolean putbackToken;

   /**
    * The symbol table - contains PascalSymbol
    */
   private PascalSymbolTable symbolTable;

   /**
    * The register table - contains PascalSymbol 
    * (of what is currently in registers)
    */
   private List registerTable;

   /**
    * The current symbol we are examining
    */
   private PascalSymbol currentSymbol;

   /**
    * The current type we are examining
    */
   private PascalType currentType;

   /**
    * The filename (if known) of the file we are compiling.
    */
   private String filename;

   /**
    * Condition variable may be set to indicate special things.
    * FIXME: This might not be the best of ideas.
    */
   private int conditionIfToken;
   private int conditionNotToken;

   /*
    * There are some conditions that can occur:
    *       Failure                 Success
    * AND   Failure of Expression   Continue evaluating
    * OR    Continue evaluating     Success of Expression
    *
    * Therefore, we need the following lists of fixups:
    * - After expression and statement block (eval to FALSE)
    * - After expression and begin of statement block (eval to TRUE)
    * - Next Expression Block (not sure yet - flushed with ORs)
    */
   private final static int FIXUP_LENGTH=2500;
   private final static int FIXUP_INVALID=0;
   private final static int FIXUP_COMP=1;
   private final static int FIXUP_IMM=2;
   private final static int FIXUP_MAY=3;
   private int[] fixups;

   /**
    * Constructor.
    *
    * Initializes internal requirements (PascalScanner and TargetArchitecture)
    */
   public PascalParser(String newFilename)
   {
      String baseFilename, outputFilename, debugFilename, assemblerFilename;

      filename = newFilename;

      int lastpos = filename.lastIndexOf(".");

      if (lastpos != -1)
         baseFilename = filename.substring(0,lastpos);
      else
         baseFilename = filename;

      outputFilename = baseFilename + ".o";
      assemblerFilename = baseFilename + ".S";
      debugFilename = baseFilename + ".dbg";

      ps = new PascalScanner();
      ta = new TargetArchitecture(outputFilename, 
                                  assemblerFilename, 
                                  debugFilename);

      typeErrorCount = 0;
      conditionIfToken = 0;
      conditionNotToken = 0;
      symbolTable = createSymbolTable();
      registerTable = new Vector();
      fixups = new int[FIXUP_LENGTH];
      //Arrays.fill(fixups,FIXUP_INVALID);
   }

   /**
    * Reads one character from the input file.
    *
    * Has backtracking ability.
    *
    * @see PascalScanner
    */   
   private int GetSym()
   {
      if (putbackToken == true)
         putbackToken = false;
      else
         ps.GetSym();

      //System.out.print(ps.currentTokenValue + "\t");

      return ps.currentToken;
   }

   /**
    * Emit predefined functions.
    */
   private void emitPredefinedFunctions()
   {
      int varOffset = 0;
      PascalType pt;
      PascalSymbol ps;

      PascalSymbol readProc = new PascalSymbol("ReadInt", 
                              PascalLanguage.KIND_PROC, 
                              ta.getInstructionCount());

      readProc.formalParameters = 1;
      symbolTable.add(readProc);
      
      pt = new PascalType(PascalLanguage.TYPE_INTEGER,
                          TargetArchitecture.SIZE_INTEGER);
      ps = new PascalSymbol("x", PascalLanguage.KIND_VAR, pt, varOffset);

      /*
       * ReadInt has a VAR parameter.
       */
      ps.setSpecial();

      readProc.addSymbol(ps);

      // we aren't really a symbol.  just grab a register.
      clearRegisters();
      loadSymbol(ps);
      ta.genOp(ta.INST_RDD, ps.registerNumber, 0, 0);
      unloadSymbol(ps);
      ta.genOp(ta.INST_RET, 0, 0, ta.REG_RET);
      clearRegisters();

      /*
      <clear registers>
      PSH 0, 29, 4
      PSH 31, 29, 4
      PSH 28, 29, 4
      BSR <address>
      <clear registers>
      POP 28, 29, 4
      POP 31, 29, 4
      POP r[free], 29, 4
      */
      /*
      LDW r[free], SP, -8
      RDD r[free]
      STW r[free], SP, -8
      RET 31
      */
      PascalSymbol writeProc = new PascalSymbol("WriteInt", 
                               PascalLanguage.KIND_PROC,
                              ta.getInstructionCount());
      ps = new PascalSymbol("x", PascalLanguage.KIND_VAR, pt, varOffset);
      writeProc.formalParameters = 1;

      symbolTable.add(writeProc);
      writeProc.addSymbol(ps);

      clearRegisters();
      loadSymbol(ps);
      ta.genOp(ta.INST_WRD, 0, ps.registerNumber, 0);
      unloadSymbol(ps);
      ta.genOp(ta.INST_RET, 0, 0, ta.REG_RET);
      clearRegisters();

      PascalSymbol writelineProc = new PascalSymbol("WriteLn", 
                                   PascalLanguage.KIND_PROC,
                                   ta.getInstructionCount());

      writelineProc.formalParameters = 0;
      symbolTable.add(writelineProc);

      ta.genOp(ta.INST_WRL, 0, 0, 0);
      ta.genOp(ta.INST_RET, 0, 0, ta.REG_RET);
   }

   /**
    * @return A new symbol table
    */
   private PascalSymbolTable createSymbolTable()
   {
      // FIXME: scopes...
      // FIXME: clone it...
      // FIXME: don't copy it, point at it.
      // FIXME: write a symbol table =)

      PascalSymbolTable l = new PascalSymbolTable();

      if (symbolTable == null)
      {
         // Predefined types only go in top level.
         // FIXME: cheese
         PascalType integerType = new PascalType(PascalLanguage.TYPE_INTEGER,
                                               TargetArchitecture.SIZE_INTEGER);

         PascalType booleanType = new PascalType(PascalLanguage.TYPE_BOOLEAN,
                                               TargetArchitecture.SIZE_BOOLEAN);

         l.add(new PascalSymbol("INTEGER", PascalLanguage.KIND_TYPE, integerType));
         l.add(new PascalSymbol("BOOLEAN", PascalLanguage.KIND_TYPE, booleanType));
         l.add(new PascalSymbol("TRUE", PascalLanguage.KIND_CONST, booleanType, 1));
         l.add(new PascalSymbol("FALSE", PascalLanguage.KIND_CONST, booleanType, 0));

      }
      else  
         // FIXME: This isn't always the case.  Only KIND_TYPE and 
         // KIND_CONST can descend down the scope of a RECORD variable.
         l.setParent(symbolTable);


      return l;
   }

   /**
    * @return New PascalType that refers to an INTEGER
    */
   private PascalType createIntegerType()
   {
      // no Reals in Pascal-2001 - create an integer
      return new PascalType(PascalLanguage.TYPE_INTEGER,
                            TargetArchitecture.SIZE_INTEGER);
   }

   private void markFixup(int loc, int type)
   {  fixups[loc] = type;  }

   private void markCompleteFixup(int loc)
   {  markFixup(loc, FIXUP_COMP);  }

   private void markCompleteFixup()
   {  markCompleteFixup(ta.getInstructionCount() - 1);   }

   private void markImmediateFixup(int loc)
   {  markFixup(loc, FIXUP_IMM);   }

   private void markImmediateFixup()
   {  markImmediateFixup(ta.getInstructionCount() - 1);  }

   private void markMayFixup(int loc)
   {  markFixup(loc, FIXUP_MAY);   }

   private void markMayFixup()
   {  markMayFixup(ta.getInstructionCount() - 1);  }

   private void clearFixups(int type)
   {
      for (int i = 0; i < fixups.length ; i++)
      {
         if (fixups[i] == type)
         {
            ta.fixupInstruction(i, ta.getInstructionCount());
            fixups[i] = FIXUP_INVALID;
         }
      }
   }

   private void clearCompleteFixups()
   { clearFixups(FIXUP_COMP); }

   private void clearImmediateFixups()
   { clearFixups(FIXUP_IMM); }

   private void clearMayFixups()
   { clearFixups(FIXUP_MAY); }

   /*
   private void printRegisterTable()
   {
      ListIterator li = registerTable.listIterator();
      while (li != null && li.hasNext())
      {
         Object o = li.next();
         if (o instanceof PascalSymbol)
         {
            PascalSymbol sym = (PascalSymbol)o;
            System.out.println(sym.registerNumber);      
         }
      }
   }
   */

   private void saveRegisters(List registerList, PascalSymbolTable excludes)
   {
      // We PSH in ascending order, POP in descending order
      ListIterator li = registerList.listIterator();
      while (li != null && li.hasNext())
      {
         Object o = li.next();
         if (o instanceof PascalSymbol)
         {
            PascalSymbol sym = (PascalSymbol)o;
            if (!excludes.contains(sym))
            {
               ta.genOp(ta.INST_PSH, sym.registerNumber, ta.REG_SP, 
                        -ta.SIZE_INTEGER);
            }
         }
      }
   }

   private void saveRegisters(List registerList)
   {
      saveRegisters(registerList, new PascalSymbolTable());
   }

   /**
    * Restores registers from the register table EXCEPT for those
    * contained within the excludes list.
    *
    * @param registerList The register table.
    * @param excludes The Exclude list for restoring from the register table.
    */
   private void restoreRegisters(List registerList, PascalSymbolTable excludes)
   {
      // We PSH in ascending order, POP in descending order
      for (int i = registerList.size() - 1; i >= 0; i--)
      {
         Object o = registerList.get(i);
         if (o instanceof PascalSymbol)
         {
            PascalSymbol sym = (PascalSymbol)o;

            if (!excludes.contains(sym))
               ta.genOp(ta.INST_POP, sym.registerNumber, ta.REG_SP, ta.SIZE_INTEGER);
            else
               ta.genOp(ta.INST_POP, 0, ta.REG_SP, ta.SIZE_INTEGER);
         }
      }
   }

   private void restoreRegisters(List registerList)
   {
      restoreRegisters(registerList, new PascalSymbolTable());
   }

   private List setLoadedSymbols(PascalSymbol[] loaded)
   {
      List registerList = new Vector();

      for (int i = 0; i < loaded.length; i++)
      {
         if (loaded[i] != null)
         {
            loaded[i].registerNumber = i;
            registerList.add(loaded[i]);
         }
      }

      return registerList;
   }

   private PascalSymbol[] getLoadedSymbols(List registerList)
   {
      PascalSymbol regs[] = new PascalSymbol[ta.REG_COUNT+1];

      ListIterator li = registerList.listIterator();

      while (li != null && li.hasNext())
      {
         Object o = li.next();
         if (o instanceof PascalSymbol)
         {
            PascalSymbol sym = (PascalSymbol)o;
            regs[sym.registerNumber] = sym;
         }
      }

      return regs;
   }

   /**
    * Goes through and invalidates ALL registers.
    *
    * NOTE: They are NOT stored back into memory.
    */
   private void invalidateRegisters(List registerList)
   {
      // Clear out everything else now.
      ListIterator li = registerList.listIterator();
      while (li != null && li.hasNext())
      {
         Object o = li.next();
         if (o instanceof PascalSymbol)
         {
            PascalSymbol sym = (PascalSymbol)o;
            if (sym.registerNumber != ta.REG_INVALID)
               ta.deallocateRegister(sym.registerNumber);
            li.remove();
         }
      }
   }

   /**
    * Allocate register for general purpose use.
    *
    * May force deallocation of other registers if full.
    *
    * This function should not be directly unless you know
    * what you are doing.
    */
   private int allocateRegister()
   {
      int registerNumber = ta.allocateRegister();

      // Registers are full - Force run reaper...
      if (registerNumber == TargetArchitecture.REG_INVALID)
      {
         clearRegisters();

         registerNumber = ta.allocateRegister();

         // we are in bad, bad, bad shape...
         // Potential unfreeing of temp registers?
         if (registerNumber == TargetArchitecture.REG_INVALID)
            throw new UnsupportedOperationException();
      }

      return registerNumber;
   }

   /**
    * Allocate register for a particular symbol.
    *
    * May force deallocation of other registers if full.
    *
    * This function should not be directly unless you know
    * what you are doing.
    *
    * @param s PascalSymbol that needs to be loaded
    */
   private void allocateRegister(PascalSymbol s)
   {
      if (s.registerNumber != TargetArchitecture.REG_INVALID)
         return;

      s.registerNumber = ta.allocateRegister();

      // Registers are full - Force run reaper...
      if (s.registerNumber == TargetArchitecture.REG_INVALID)
      {
         clearRegisters();

         s.registerNumber = ta.allocateRegister();

         // we are in bad, bad, bad shape...
         if (s.registerNumber == TargetArchitecture.REG_INVALID)
            throw new UnsupportedOperationException();
      }

      registerTable.add(s);
   }

   /**
    * Swap the two symbols in the register table.
    *
    * @param l Left symbol
    * @param r Right symbol
    */
   private void swapRegisters(PascalSymbol l, PascalSymbol r)
   {
      // nothing to do...
      if (l.registerNumber == TargetArchitecture.REG_INVALID &&
         r.registerNumber == TargetArchitecture.REG_INVALID)
         return;

      int lReg = l.registerNumber;
      int rReg = r.registerNumber;
   
      // if both are valid, they both need to stay in table   
      if (lReg == TargetArchitecture.REG_INVALID)
      {
         registerTable.remove(r);
         registerTable.add(l);
      }
      else if (rReg == TargetArchitecture.REG_INVALID)
      {
         registerTable.remove(l);
         registerTable.add(r);
      }

      l.registerNumber = rReg;
      r.registerNumber = lReg;
   }

   /**
    * This function takes the register assigned to the right
    * symbol and gives it to the left one.
    *
    * In cases where the right symbol is tied to a named value,
    * it must be written as it no longer controls the value.
    *
    * Optimizations could be made here to delay this until
    * the left symbol is modified by another operation, then
    * you can proceed to store the old value into the right
    * symbol, and then overwrite the new one.
    *
    * @param l Left symbol
    * @param r Right symbol
    */
   private void switchRegister(PascalSymbol l, PascalSymbol r)
   {
      int lReg = l.registerNumber;
      int rReg = r.registerNumber;
   
      // the right symbol isn't loaded, load it into left directly.
      if (rReg == TargetArchitecture.REG_INVALID)
      {
         // AllocateRegister MAY give a register or use the old value.
         allocateRegister(l);
         readSymbol(r, l.registerNumber, 0);
      }
      else
      {
         // if the left one isn't in, add it.
         // we might be overriding an old value
         if (lReg == TargetArchitecture.REG_INVALID)
            registerTable.add(l);
         else
            ta.deallocateRegister(l.registerNumber);

         // the unload function will also deallocate the register
         // we just want to remove the named symbols's link in the 
         // register table and store it's value.
         if (r.isNamed() == true && r.kind == PascalLanguage.KIND_VAR)
         {
            writeSymbol(r, rReg, 0);

            r.registerNumber = ta.REG_INVALID;
         }

         registerTable.remove(r);
         l.registerNumber = rReg;
      }
   }

   private void copyData(PascalSymbol left, PascalType leftType,
                         PascalSymbol right, PascalType rightType)
      throws ParsingException
   { 
      int regL, regR, regA;

      if (leftType.compareTo(rightType) != 0)
         throw new TypeException(left, right);

      switch (leftType.form)
      {
      case PascalLanguage.TYPE_INTEGER:
      case PascalLanguage.TYPE_BOOLEAN:
         switchRegister(left, right);
         break;
      case PascalLanguage.TYPE_ARRAY:
         unloadSymbol(right);
         regL = allocateRegister();
         regR = allocateRegister();
         regA = allocateRegister();
         ta.genOp(ta.INST_ADDI, regL, 0, left.address);
         ta.genOp(ta.INST_ADDI, regR, 0, right.address);
         ta.genOp(ta.INST_LDX,  regA, ta.getVarReg(), regR);
         ta.genOp(ta.INST_STX,  regA, ta.getVarReg(), regL);
         ta.genOp(ta.INST_SUBI, regL, regL, leftType.base.sizeOf);
         ta.genOp(ta.INST_SUBI, regR, regR, leftType.base.sizeOf);
         ta.genOp(ta.INST_CMPI, regA, regL, left.address - leftType.sizeOf);
         ta.genOp(ta.INST_BGT,  regA, 0, -5);
         ta.deallocateRegister(regL);
         ta.deallocateRegister(regR);
         ta.deallocateRegister(regA);
         break;
      case PascalLanguage.TYPE_RECORD:
         unloadSymbol(right);
         regA = allocateRegister();
         List symbols = leftType.symbolTable;
         ListIterator li = symbols.listIterator();
         while (li != null && li.hasNext())
         {
            Object o = li.next();
            if (o instanceof PascalSymbol)
            {
               PascalSymbol ps = (PascalSymbol)o;
               readSymbol(right, regA, ps.address);
               writeSymbol(left, regA, ps.address);
            }
         }
         ta.deallocateRegister(regA);
         break;
      default:
         throw new TypeException(left, right);
      }
   }

   /**
    * Stores symbol contents back in memory
    */
   private void writeSymbol(PascalSymbol s, int reg, int offset)
   {
      if (reg == ta.REG_INVALID)
         return;

      int storeAddress = s.address - offset;

      int baseReg = getBasePointer(symbolTable, s);

      switch (s.type.form)
      {
      case PascalLanguage.TYPE_ARRAY:
      case PascalLanguage.TYPE_RECORD:
         // s.type.link == null if constant index
         // r.type.link != null if variable-based index
         if (s.type.link == null)
         {
            storeAddress -= s.addressOffset;
            ta.genOp(ta.INST_STW, reg, baseReg, storeAddress);
         }
         else
            ta.genOp(ta.INST_STX, reg, baseReg, s.type.link.registerNumber );
         break;
      case PascalLanguage.TYPE_INTEGER:
         ta.genOp(ta.INST_STW, reg, baseReg, storeAddress);
         break;
      default:
         throw new UnsupportedOperationException();
      }

      clearBasePointer(baseReg);
   }

   private void loadHugeConstant(PascalSymbol s, int reg)
   {
      int MAX_SPLIT = 3;

      int split, rem, bl, len;
      int j[] =  new int[MAX_SPLIT+1];
      int i = s.address;
 
      String number = Integer.toBinaryString(i);
      String temp;

      len = number.length();
      split = len / MAX_SPLIT;
      rem = len % MAX_SPLIT;

      for (int pos = 0; pos <= MAX_SPLIT; pos++)
      {
         len = number.length();
         
         if (number.length() == 0)
            break;

         if (split <= len)
         {
            temp = number.substring(0,split); 
            number = number.substring(split); 
         }
         else
            temp = number;

         j[pos] = Integer.parseInt(temp, 2);
      }
   
      int k = 0; 
      bl = 0;

      ta.genOp(ta.INST_ADDI, reg, 0, 0);

      do
      {
         ta.genOp(ta.INST_LSHI, reg, reg, split);
         ta.genOp(ta.INST_ADDI, reg, reg, j[k++]);
      } 
      while (k < MAX_SPLIT);

      if (rem > 0)
      {
         ta.genOp(ta.INST_LSHI, reg, reg, rem);
         ta.genOp(ta.INST_ADDI, reg, reg, j[k]);
      }
   }

   private int getBasePointer(PascalSymbolTable pst, PascalSymbol s)
   {
      int baseReg = ta.getVarReg();
      int newReg = 0;
      int ourDepth = pst.getSymbolTableDepth();
      int symDepth = pst.getSymbolDepth(s);

      if (symDepth != 0)
      {
         if (symDepth < 0)
         {
            if (ourDepth > 0)
               baseReg = ta.REG_GLOBAL;
         }
         else
         {
            newReg = allocateRegister();
            ta.genOp(ta.INST_LDW, newReg, baseReg, -4 * (ourDepth - symDepth+1));
            baseReg = newReg;
         } 
      }

      return baseReg;
   }

   private void clearBasePointer(int reg)
   {
      if (reg != ta.REG_GLOBAL && reg != ta.REG_FP)
         ta.deallocateRegister(reg);
   }

   private void readSymbol(PascalSymbol s, int reg, int offset)
   {
      switch (s.kind)
      {
      case PascalLanguage.KIND_VAR:
         // array offset support
         int loadAddress = s.address - offset;
         int baseReg = getBasePointer(symbolTable, s);

         // our addressing scheme is based on negative values.
         switch (s.type.form)
         {
         case PascalLanguage.TYPE_ARRAY:
         case PascalLanguage.TYPE_RECORD:
            if (s.type.link == null)
            {
               loadAddress -= s.addressOffset;
               ta.genOp(ta.INST_LDW, reg, baseReg, loadAddress);
            }
            else
               ta.genOp(ta.INST_LDX, reg, baseReg, 
                        s.type.link.registerNumber);
            break;
         case PascalLanguage.TYPE_INTEGER:
         case PascalLanguage.TYPE_BOOLEAN:
            ta.genOp(ta.INST_LDW, reg, baseReg, loadAddress);
            break;
         default:
            throw new UnsupportedOperationException(s.toString());
         }
         clearBasePointer(baseReg);
         break;
      case PascalLanguage.KIND_CONST:
         ta.genOp(ta.INST_ADDI, reg, 0, s.address);
         break;
      case PascalLanguage.KIND_HUGECONST:
         loadHugeConstant(s, reg);
         break;
      default:
         throw new UnsupportedOperationException(s.toString());
      }
   }

   private void pushSymbol(PascalItem i)
   {
      int regInt, regIndex, loadAddress;

      switch (i.type.form)
      {
      case PascalLanguage.TYPE_INTEGER:
      case PascalLanguage.TYPE_BOOLEAN:
         loadSymbol(i.symbol); 
         ta.genOp(ta.INST_PSH, i.symbol.registerNumber, ta.REG_SP, 
                  -i.type.sizeOf);
         break;
      case PascalLanguage.TYPE_ARRAY:
      case PascalLanguage.TYPE_RECORD:
         // Be cheesy and copy everything.
         loadAddress = i.symbol.address; 
         regInt = allocateRegister();
         regIndex = allocateRegister();
         if (i.symbol.type.link == null)
         {
            loadAddress -= i.symbol.addressOffset;
            ta.genOp(ta.INST_ADDI, regIndex, 0, loadAddress);
         }
         else
            ta.genOp(ta.INST_ADD, regIndex, i.symbol.type.link.registerNumber, 0);
         ta.genOp(ta.INST_LDX,  regInt, ta.getVarReg(), regIndex);
         ta.genOp(ta.INST_PSH,  regInt, ta.REG_SP, -ta.SIZE_INTEGER);
         ta.genOp(ta.INST_SUBI, regIndex, regIndex, ta.SIZE_INTEGER);
         ta.genOp(ta.INST_CMPI, regInt, regIndex, i.symbol.address - i.type.sizeOf);
         ta.genOp(ta.INST_BGT,  regInt, 0, -4);
         ta.deallocateRegister(regInt);
         ta.deallocateRegister(regIndex);
         break;
      default:
         throw new UnsupportedOperationException();
      }
   }

   private void popSymbol(PascalItem i)
   {
      int regInt, regIndex, loadAddress;

      switch (i.type.form)
      {
      case PascalLanguage.TYPE_INTEGER:
      case PascalLanguage.TYPE_BOOLEAN:
         if (i.symbol.registerNumber != ta.REG_INVALID)
         {
            ta.genOp(ta.INST_POP, i.symbol.registerNumber, ta.REG_SP, 
                     i.type.sizeOf);
         }
         else
         { 
            int tempReg = allocateRegister();
            ta.genOp(ta.INST_POP, tempReg, ta.REG_SP, i.type.sizeOf);
            writeSymbol(i.symbol, tempReg, 0); 
            ta.deallocateRegister(tempReg);
         }
         break;
      case PascalLanguage.TYPE_ARRAY:
      case PascalLanguage.TYPE_RECORD:
         // Be cheesy and copy everything.
         loadAddress = i.symbol.address - i.type.sizeOf + ta.SIZE_INTEGER;
         regInt = allocateRegister();
         regIndex = allocateRegister();
         if (i.symbol.type.link == null)
         {
            loadAddress -= i.symbol.addressOffset;
            ta.genOp(ta.INST_ADDI, regIndex, 0, loadAddress);
         }
         else
            ta.genOp(ta.INST_SUBI, regIndex, i.symbol.type.link.registerNumber, i.type.sizeOf + ta.SIZE_INTEGER);
         ta.genOp(ta.INST_POP,  regInt, ta.REG_SP, ta.SIZE_INTEGER);
         ta.genOp(ta.INST_STX,  regInt, ta.getVarReg(), regIndex);
         ta.genOp(ta.INST_ADDI, regIndex, regIndex, ta.SIZE_INTEGER);
         ta.genOp(ta.INST_CMPI, regInt, regIndex, i.symbol.address);
         ta.genOp(ta.INST_BLE,  regInt, 0, -4);
         ta.deallocateRegister(regInt);
         ta.deallocateRegister(regIndex);
         break;
      default:
         throw new UnsupportedOperationException();
      }
   }

   /**
    * Slightly simplified version that allows them to be discarded.
    */
   private void popSymbol(PascalItem i, boolean clear)
   {
      if (!clear)
      {
         popSymbol(i);
         return;
      }

      int regInt, regIndex, loadAddress;

      switch (i.type.form)
      {
      case PascalLanguage.TYPE_INTEGER:
      case PascalLanguage.TYPE_BOOLEAN:
         ta.genOp(ta.INST_POP, 0, ta.REG_SP, i.type.sizeOf);
         break;
      case PascalLanguage.TYPE_ARRAY:
      case PascalLanguage.TYPE_RECORD:
         // Be cheesy and copy everything.
         // FIXME: POP with a high SIZE value?
         loadAddress = i.symbol.address - i.type.sizeOf;
         regIndex = allocateRegister();
         regInt = allocateRegister();
         if (i.symbol.type.link == null)
         {
            loadAddress -= i.symbol.addressOffset;
            ta.genOp(ta.INST_ADDI, regIndex, 0, loadAddress);
         }
         else
            ta.genOp(ta.INST_SUBI, regIndex, i.symbol.type.link.registerNumber, i.type.sizeOf);
         ta.genOp(ta.INST_POP,  0, ta.REG_SP, ta.SIZE_INTEGER);
         ta.genOp(ta.INST_ADDI, regIndex, regIndex, ta.SIZE_INTEGER);
         ta.genOp(ta.INST_CMPI, regInt, regIndex, i.symbol.address);
         ta.genOp(ta.INST_BLE,  regInt, 0, -3);
         ta.deallocateRegister(regInt);
         ta.deallocateRegister(regIndex);
         break;
      default:
         throw new UnsupportedOperationException();
      }
   }

   /**
    * Loads generic symbol into memory.
    *
    * @param s Symbol to be loaded
    */
   private void loadSymbol(PascalSymbol s)
   {
      // already loaded
      if (s.registerNumber != ta.REG_INVALID)
         return;

      allocateRegister(s);

      readSymbol(s, s.registerNumber, 0);
   }

   /**
    * Unloads a generic symbol from memory.
    *
    * Named symbols need to be stored when this occurs.
    * Unnamed symbols can be silently discarded.
    *
    * NOTE: IF YOU CALL THIS FUNCTION, YOU MUST
    * DELETE IT FROM registerTable!
    *
    * @param s Symbol to be unloaded
    * @see unloadSymbol
    */
   private void discardSymbol(PascalSymbol s)
   {
      // If we loaded it somehow, then we will unload it
      // without question
      if (s.registerNumber == ta.REG_INVALID)
         return;

      if (s.isNamed() == true && s.kind == PascalLanguage.KIND_VAR)
         writeSymbol(s, s.registerNumber, 0);

      // If there was an index register, we are done with it now.
      // The reaper should remove it when it can.
      if (s.type != null && s.type.link != null)
         s.type.link = null;

      ta.deallocateRegister(s.registerNumber);

      s.registerNumber = ta.REG_INVALID;
   }

   /**
    * Removes and frees up a register.
    *
    * @see discardSymbol
    */
   private void unloadSymbol(PascalSymbol s)
   {
      discardSymbol(s);
      registerTable.remove(s);
   }

   /**
    * Goes through all registers and unloads all of them.
    */
   private void clearRegisters()
   {
      if (registerTable != null)
      {  
         // due to KIND_LINK types, we MUST
         // unload named registers first. 
         ListIterator li = registerTable.listIterator();
         while (li != null && li.hasNext())
         {
            Object o = li.next();
            if (o instanceof PascalSymbol)
            {
               PascalSymbol sym = (PascalSymbol)o;
               if (sym.isNamed() == true)
               {
                  discardSymbol(sym);
                  li.remove();
               }
            }
         }

         // Clear out everything else now.
         li = registerTable.listIterator();
         while (li != null && li.hasNext())
         {
            Object o = li.next();
            if (o instanceof PascalSymbol)
            {
               PascalSymbol sym = (PascalSymbol)o;
               discardSymbol(sym);
               li.remove();
            }
         }
      }
   }

   /**
    * Goes through all registers and clears all unnamed symbols.
    *
    * Good for relieving register pressure from temporary symbols.
    */
   private void clearUnnamedRegisters()
   {
      if (registerTable != null)
      {   
         ListIterator li = registerTable.listIterator();
         while (li != null && li.hasNext())
         {
            Object o = li.next();
            if (o instanceof PascalSymbol)
            {
               PascalSymbol sym = (PascalSymbol)o;
               if ((!sym.isNamed() && !sym.isLinked()) || sym.isSpecial())
               {
                  discardSymbol(sym);
                  li.remove();
               }
            }
         }
      }
   }

   /**
    * This function will handle loading of symbols and
       * optimizing constants.
    *
    * FIXME: It will not handle unloading of symbols - yet.
    *
    * @param opCode   The operation wish to be performed (no immediates!)
    * @param l    Left symbol
    * @param r    Right symbol
    *
    * @return The returned symbol.
    */
   private PascalSymbol optimizeSymbols(int opCode, 
                    PascalSymbol l, PascalSymbol r)
   {
      PascalSymbol s = new PascalSymbol(PascalLanguage.KIND_VAR);   

      // yes, the memory addresses may be equivalent   
      if (l != null && l != r && 
          (l.kind == PascalLanguage.KIND_CONST ||
           l.kind == PascalLanguage.KIND_HUGECONST) &&
          (r.kind == PascalLanguage.KIND_CONST ||
           r.kind == PascalLanguage.KIND_HUGECONST) &&
          l.type.compareTo(r.type) == 0)
      {
         // we need to now override the kind
         s.kind = PascalLanguage.KIND_CONST;

         // Assume the types are identical
         s.type = l.type;

         // These are our constant optimizers.
         switch (opCode)
         {
         case TargetArchitecture.INST_ADD:
            s.address = l.address + r.address;
            break;
         case TargetArchitecture.INST_SUB:
            s.address = l.address - r.address;
            break;
         case TargetArchitecture.INST_MUL:
            s.address = l.address * r.address;
            break;
         case TargetArchitecture.INST_DIV:
            s.address = l.address / r.address;
            break;
         case TargetArchitecture.INST_MOD:
            s.address = l.address % r.address;
            //System.out.println(s.address + "=" + l.address + "%" + r.address);
            break;
         case TargetArchitecture.INST_CMP:
            s.address = (l.address == r.address) ? 1 : 0;
            break;
         case TargetArchitecture.INST_OR:
            // FIXME: Verify these
            s.address = l.address | r.address;
            break;
         case TargetArchitecture.INST_AND:
            // FIXME: Verify these
            s.address = l.address & r.address;
            break;
         case TargetArchitecture.INST_BIC:
            // FIXME: Verify these
            s.address = l.address & r.address;
            break;
         case TargetArchitecture.INST_XOR:
            // FIXME: Verify these
            s.address = l.address ^ r.address;
            break;
         case TargetArchitecture.INST_LSH:
            // FIXME: Verify these
            s.address = l.address << r.address;
            break;
         case TargetArchitecture.INST_RSH:
            // FIXME: Verify these
            s.address = l.address >> r.address;
            break;
         case TargetArchitecture.INST_CHK:
            // FIXME: Verify these
            s.address = (l.address == r.address) ? 1 : 0;
            break;
         default:
            throw new UnsupportedOperationException();
         }
           
         if (s.address > ta.MAX_INT)
            s.kind = PascalLanguage.KIND_HUGECONST;
      }
      else
      {
         // In the special case of a unary symbol (+<ident>),
         // we fake the left symbol by temporarily creating
         // a new symbol and assigning it register 0.
         // it will fall out of scope almost immediately.
         if (l == null)
         {
            l = new PascalSymbol();

            // yes.
            l.registerNumber = 0;

            // for below.
            l.type = r.type;
         }
         // Otherwise, insure that the symbol is loaded
         else if (l.registerNumber == TargetArchitecture.REG_INVALID) 
            loadSymbol(l);
      
         allocateRegister(s);   
         s.type = l.type;

         if (r.kind == PascalLanguage.KIND_CONST)
            ta.genOp(opCode + 16, s.registerNumber, l.registerNumber, r.address);
         else
         {
            if (r.registerNumber == TargetArchitecture.REG_INVALID) 
               loadSymbol(r);
      
            ta.genOp(opCode, s.registerNumber, l.registerNumber, r.registerNumber); 
         }

         // To be faithful to the doctrines given by
         // Assignment 7, we should always unload
         // the symbols after we are done.
         // Instead, we delay this until we are either
         // out of registers or we have exited the
         // statement sequence.
         // Per discussion with Dr. Franz, this is fine.
         //unloadSymbol(l);   
         //unloadSymbol(r);   
      }

      return s;   
   }

   private void parseIdentifier(int token)
      throws ParsingException
   {
      if (token != PascalLanguage.TOKEN_IDENT)
         throw new SyntaxException(token, PascalLanguage.TOKEN_IDENT);
   }

   private void parseNumber(int token)
      throws ParsingException
   {
      if (token != PascalLanguage.TOKEN_NUMBER)
         throw new SyntaxException(token, PascalLanguage.TOKEN_NUMBER);
   }

   /**
    * This function handles arrays and resolving the correct
    * type of the variable.
    *
    */
   private void parseSelector(int token)
      throws ParsingException
   {
      // In a desperate attempt to make the code
      // slightly cleaner, this statement was added.
      if (token != PascalLanguage.TOKEN_LB &&
          token != PascalLanguage.TOKEN_PERIOD)
      {
         currentType = currentSymbol.type;
         putbackToken = true;
         return;
      }

      // We can not store anything in the current symbol
      // until we are done resolving the symbol entirely
      // because we may be recursive in nature.
      PascalSymbol saveSym = currentSymbol;

      // If any of the indices are variables,
      // we have to load the index into a register
      // and add the constant values.
      // Otherwise, we handle this at compile time.  

      // We will store four things:
      // The previous address we loaded from
      // The previous "Link" variable we used
      // Current address to load from
      // Current "Link" variable we used
      int lastAddress = saveSym.addressOffset;
      PascalSymbol lastField = saveSym.type.link;
      int currentAddress = 0;
      PascalSymbol currentField = null;

      // this will hold our type as we descend down the parse tree
      PascalType curType = saveSym.type;
   
      int currentToken = token;

      // we can guarantee that the first token is correct,
      // so save evaluating it for properness.
      do
      {
         // Array is indicated by TOKEN_LB
         // Record is indicated by TOKEN_PERIOD
         switch (currentToken)
         {
         case PascalLanguage.TOKEN_LB:
            // we have gone too far
            if (curType.form != PascalLanguage.TYPE_ARRAY)
               throw new TypeException(saveSym);
  
            parseExpression(GetSym());

            if (GetSym() != PascalLanguage.TOKEN_RB)
               throw new SyntaxException(ps.currentToken, PascalLanguage.TOKEN_RB);

            // currentSymbol now points to the identifier.
            // either it is a constant or a variable.
            switch (currentSymbol.kind)
            {
            case PascalLanguage.KIND_CONST:
               if (currentSymbol.address < 0 || currentSymbol.address > curType.length)
                  throw new TypeException(saveSym);
            
               // advance the offset by the constant value
               currentAddress += curType.base.sizeOf * currentSymbol.address; 
               break;
            case PascalLanguage.KIND_VAR:
               // we should already have it loaded into a register
               if (currentSymbol.registerNumber == ta.REG_INVALID)
                  loadSymbol(currentSymbol);

               // execute the dynamic instruction for bounds checking
               ta.genOp(ta.INST_CHKI, currentSymbol.registerNumber, 0, curType.length);

               // FIXME: This is quasi-sub-optimal because we could 
               // end up doing a check for multiple indices which would
               // eventually be equivalent.
               // Data-Flow Analysis would need to be done to
               // determine if the eventual index hasn't changed
               // since the last time the array was accessed.
               // If we can guarantee it hasn't changed (only after
               // going through all indices), then we don't have to
               // do anything.  For now, just allocate a reg.
               if (currentField == null)
               {
                  // This is a symbol which is linked to the array.
                  currentField = new PascalSymbol(PascalLanguage.KIND_LINK);

                  allocateRegister(currentField);

                  // account for base value which is negative
                  ta.genOp(ta.INST_ADDI, currentField.registerNumber, 
                           0, saveSym.address );
               }
   
               // We are going to allocate the register and deallocate it here.
               int tempReg = allocateRegister();

               // account for the size of our type
               ta.genOp(ta.INST_MULI, tempReg, 
                        currentSymbol.registerNumber, curType.base.sizeOf);

               // subtract the correct value to our offset marker
               ta.genOp(ta.INST_SUB, currentField.registerNumber, 
                        currentField.registerNumber, tempReg);
               
               ta.deallocateRegister(tempReg);
               break;
            default:
               throw new TypeException(currentSymbol);
            }

            // advance the local type
            curType = curType.base;
            break;
         case PascalLanguage.TOKEN_PERIOD:
            // Record

            if (curType.form != PascalLanguage.TYPE_RECORD)
               throw new TypeException(currentSymbol, curType);
 
            // load the records symbol table and find it.
            PascalSymbolTable oldSymbolTable = symbolTable;
            symbolTable = curType.symbolTable;
       
            parseSymbol(GetSym());

            // FIXME: Symbols with TYPE_LINK?
            // FIXME: Verify addressOffset
            //currentAddress += currentSymbol.address + currentSymbol.addressOffset;
            currentAddress += currentSymbol.address;
        
            // Restore old parameters
            symbolTable = oldSymbolTable;
            currentSymbol = saveSym;

            // when we call parseSymbol(), currentType is set to 
            // the type of the record
            curType = currentType;
            break;
         default:
               throw new TypeException(currentSymbol);
         }

         currentToken = GetSym();
      }
      while (currentToken == PascalLanguage.TOKEN_LB ||
             currentToken == PascalLanguage.TOKEN_PERIOD);

      // Now, we are done with the selector.

      // always rollback one
      putbackToken = true;

      // set the global type to what we last were
      currentType = curType;

      // restore the symbol to what we were originally
      currentSymbol = saveSym;

      // if it is a constant, type.link is never set.
      if (currentField == null)
      {
         // if it isn't the same address as before
         if (lastAddress != currentAddress || currentSymbol.type.link != null)
         {
            // force it to be cleared from register
            unloadSymbol(currentSymbol);

            currentSymbol.addressOffset = currentAddress;
            currentSymbol.type.link = null;
          }
      }
      else // it is loaded in a register
      {
         // if we have constant section, add it to our register
         if (currentAddress != 0)
            ta.genOp(ta.INST_SUBI, currentField.registerNumber, 0, currentAddress);

         // detach old reference
         unloadSymbol(currentSymbol);

         // now attach it
         currentSymbol.addressOffset = 0;
         currentSymbol.type.link = currentField;
      }

   }

   private void parseSymbol(int token)
      throws ParsingException
   {
      parseIdentifier(token);

      String currentToken = ps.currentTokenValue;

      currentSymbol = symbolTable.getSymbol(currentToken, 
                                 PascalLanguage.KIND_VAR);

      // could also be a constant.
      if (currentSymbol == null)
         currentSymbol = symbolTable.getSymbol(currentToken, 
                    PascalLanguage.KIND_CONST);

      // could also be a huge constant.
      if (currentSymbol == null)
         currentSymbol = symbolTable.getSymbol(currentToken, 
                    PascalLanguage.KIND_HUGECONST);

      // could also be a procedure.
      if (currentSymbol == null)
         currentSymbol = symbolTable.getSymbol(currentToken,
                           PascalLanguage.KIND_PROC);

      if (currentSymbol == null)
         throw new TypeException(currentToken);

      currentType = currentSymbol.type;

      parseSelector(GetSym());
   }

   private void parseFactor(int token)
      throws ParsingException
   {
      if (token == PascalLanguage.TOKEN_IDENT)
      {
         parseSymbol(token);
      }
      else if (token == PascalLanguage.TOKEN_NUMBER)
      {
         parseNumber(token);

         currentType = createIntegerType();

         currentSymbol = new PascalSymbol(PascalLanguage.KIND_CONST,
                   currentType,
                   Integer.parseInt(ps.currentTokenValue));

      }
      else if (token == PascalLanguage.TOKEN_LP)
      {
         parseExpression(GetSym());
         if (GetSym() != PascalLanguage.TOKEN_RP)
            throw new SyntaxException(ps.currentToken, PascalLanguage.TOKEN_RP);
      }
      else if (token == PascalLanguage.TOKEN_NOT)
      {
         // Let's note this condition and continue.
         int oldConditionToken = conditionNotToken;

         // special case
         if (conditionNotToken == PascalLanguage.TOKEN_NOT)
            conditionNotToken = PascalLanguage.TOKEN_INVALID;
         else
            conditionNotToken = token;

         parseFactor(GetSym()); 

         // restore
         conditionNotToken = oldConditionToken;
      }

   }

   private void parseTerm(int token)
      throws ParsingException
   {
      parseFactor(token);

      int saveInstruction = ta.getInstructionCount() - 1;

      int currentToken = GetSym();

      if (currentToken == PascalLanguage.TOKEN_STAR ||
         currentToken == PascalLanguage.TOKEN_DIV ||
         currentToken == PascalLanguage.TOKEN_MOD ||
         currentToken == PascalLanguage.TOKEN_AND)
      {
         // can only be Integers when we are doing math
         if (currentType.form != PascalLanguage.TYPE_INTEGER)
            throw new TypeException(currentSymbol);

         PascalType leftType = currentType;
         PascalSymbol leftSym = currentSymbol;

         // optional, but can occur infinite times
         while (currentToken == PascalLanguage.TOKEN_STAR ||
            currentToken == PascalLanguage.TOKEN_DIV ||
            currentToken == PascalLanguage.TOKEN_MOD ||
            currentToken == PascalLanguage.TOKEN_AND)
         {
            parseFactor(GetSym());

            if (leftType.compareTo(currentType) != 0)
               throw new TypeException(leftSym, currentType);

            int inst = 0;

            switch (currentToken)
            {
            case PascalLanguage.TOKEN_STAR:
               inst = ta.INST_MUL;
               break;
            case PascalLanguage.TOKEN_DIV:
               inst = ta.INST_DIV;
               break;
            case PascalLanguage.TOKEN_MOD:
               inst = ta.INST_MOD;
               break;
            case PascalLanguage.TOKEN_AND:
               if (conditionIfToken != PascalLanguage.TOKEN_IF)
                  inst = ta.INST_AND;
               else
               {
                  if (conditionNotToken == PascalLanguage.TOKEN_NOT)
                  {
                     ta.reverseComparisonInstruction(saveInstruction);
                     // No.  The second instruction doesn't get reversed.
                     // Don't ask me why. 
                     // FIXME: Yes, do.
                     //ta.reverseComparisonInstruction(ta.getInstructionCount() - 1);
                     //markImmediateFixup(saveInstruction);
                     //markImmediateFixup();
                  } 
                  markMayFixup(saveInstruction);
                  markImmediateFixup();
                  inst = ta.INST_INVALID;
               }
               break;
            }

            if (inst != ta.INST_INVALID)
               currentSymbol = optimizeSymbols(inst, leftSym, currentSymbol);

            leftSym = currentSymbol;
            leftType = leftSym.type;
            
            currentToken = GetSym();
         }   
         
      }

      // we ended!
      putbackToken = true;
      
   }
   
   private void parseRead(int token)
      throws ParsingException
   {
      if (GetSym() != PascalLanguage.TOKEN_LP &&
          GetSym() != PascalLanguage.TOKEN_RP)
         throw new SyntaxException(ps.currentToken, 
                    PascalLanguage.TOKEN_LP);

      currentType = createIntegerType();

      currentSymbol = new PascalSymbol(PascalLanguage.KIND_VAR,
                    currentType,
                    0);

      // we aren't really a symbol.  just grab a register.
      allocateRegister(currentSymbol);

      ta.genOp(ta.INST_RDD, currentSymbol.registerNumber, 0, 0);

      // eat the next symbol
      GetSym();
   }

   private void parseWrite(int token)
      throws ParsingException
   {
      if (GetSym() != PascalLanguage.TOKEN_LP)
         throw new SyntaxException(ps.currentToken, PascalLanguage.TOKEN_LP);
   
      parseExpression(GetSym());

      if (GetSym() != PascalLanguage.TOKEN_RP)
         throw new SyntaxException(ps.currentToken, PascalLanguage.TOKEN_RP);

      loadSymbol(currentSymbol);

      ta.genOp(ta.INST_WRD, 0, currentSymbol.registerNumber, 0);

   }

   private void parseWriteln(int token)
      throws ParsingException
   {
      int inst = ta.genOp(ta.INST_WRL, 0, 0, 0);
   }

   private void parseSimpleExpression(int token)
      throws ParsingException
   {
      int currentToken = token;

      // Expression can be unary.
      boolean unaryExpression = false;

      // OR support
      int lastInst = -1;

      // optional 
      if (currentToken != PascalLanguage.TOKEN_READ)
      {
         PascalType leftType = null;
         PascalSymbol leftSymbol = null;
     
         // the first token we see is a + or - 
         if (currentToken == PascalLanguage.TOKEN_PLUS ||
             currentToken == PascalLanguage.TOKEN_MINUS)
            unaryExpression = true;

         if (currentToken == PascalLanguage.TOKEN_OR)
            throw new SyntaxException(currentToken, PascalLanguage.TOKEN_PLUS);

         do
         {
            int lastToken = currentToken;

            // we have noted it, so it is okay to proceed
            if (currentToken == PascalLanguage.TOKEN_PLUS ||
                currentToken == PascalLanguage.TOKEN_MINUS ||
                currentToken == PascalLanguage.TOKEN_OR)
               currentToken = GetSym();

            // can't be executed first time...
            // We need to reverse the last comparison
            // so that on success it branches to success. 
            // On failure, it will continue on evaluating.
            if (lastToken == PascalLanguage.TOKEN_OR)
            {
               clearMayFixups();
               ta.reverseComparisonInstruction(lastInst);
               markImmediateFixup(lastInst);
            }
               
            parseTerm(currentToken);

            lastInst = ta.getInstructionCount() - 1;

            if (leftType == null)
            {
               if (unaryExpression == true)
                  unaryExpression = false;
               else
               {
                  leftSymbol = currentSymbol;
                  leftType = currentType;
               }
            }
            else
            {
               if (leftType.compareTo(currentType) != 0)
                  throw new TypeException(leftSymbol, currentType);
            }

            switch (lastToken)
            {
            case PascalLanguage.TOKEN_PLUS:
               currentSymbol = optimizeSymbols(ta.INST_ADD, leftSymbol, currentSymbol);
               leftSymbol = currentSymbol;
               leftType = leftSymbol.type;
               break;
            case PascalLanguage.TOKEN_MINUS:
               currentSymbol = optimizeSymbols(ta.INST_SUB, leftSymbol, currentSymbol);
               leftSymbol = currentSymbol;
               leftType = leftSymbol.type;
               break;
            case PascalLanguage.TOKEN_OR:
            }

            currentToken = GetSym();
         }   
         while (currentToken == PascalLanguage.TOKEN_PLUS || 
               currentToken == PascalLanguage.TOKEN_MINUS ||
               currentToken == PascalLanguage.TOKEN_OR);

         // we ended
         putbackToken = true;
      }
      else if (token == PascalLanguage.TOKEN_READ)
      {
         parseRead(token);
      }

   }

   private PascalType parseType(int token)
      throws ParsingException
   {
      PascalType pt = new PascalType();

      switch(token)
      {
      case PascalLanguage.TOKEN_IDENT:
      {
         parseIdentifier(token);

         PascalSymbol tempSym = symbolTable.getSymbol(ps.currentTokenValue,
                                                PascalLanguage.KIND_TYPE);
         if (tempSym == null)
            throw new TypeException(currentSymbol);
           
         pt = tempSym.type; 
         //System.out.println("looked up: " + ps.currentTokenValue + "\nfound: " + pt);
      }
      break;
      case PascalLanguage.TOKEN_ARRAY:
      {
         parseNumber(GetSym());

         pt.form = PascalLanguage.TYPE_ARRAY;
         // GRRR...
         pt.length = Integer.parseInt(ps.currentTokenValue) + 1;
            
         if (GetSym() != PascalLanguage.TOKEN_OF)
            throw new SyntaxException(token, PascalLanguage.TOKEN_OF);

         // this is recursive and will stop when it is isn't
         // an array.         
         pt.base = parseType(GetSym());

         if (pt.base != null)
            pt.sizeOf = pt.base.sizeOf * pt.length;
      }
      break;
      case PascalLanguage.TOKEN_RECORD:
      {
         int currentToken;

         pt.form = PascalLanguage.TYPE_RECORD;

         // We are going to play tricks with the symbol table.
         // Replace it with the record's symbol table.
         pt.symbolTable = createSymbolTable();
         PascalSymbolTable origSymbolTable = symbolTable;
         symbolTable = pt.symbolTable;

         // currentSymbol can be trounced by parseVars
         PascalSymbol saveSymbol = currentSymbol;

         do
         {
            currentToken = GetSym();

            if (currentToken == PascalLanguage.TOKEN_END)
               break;

            parseVars(currentToken, false); 

            currentToken = GetSym();
         }
         while (currentToken == PascalLanguage.TOKEN_SEMICOLON);

         pt.sizeOf = symbolTable.generateAddresses();

         // restore the symbol table
         symbolTable = origSymbolTable;

         // restore the original symbol
         currentSymbol = saveSymbol;

         if (ps.currentToken != PascalLanguage.TOKEN_END)
            throw new SyntaxException(ps.currentToken, PascalLanguage.TOKEN_END);
      }
      break;
      default:
         throw new SyntaxException(token, PascalLanguage.TOKEN_IDENT);
      }

      return pt;
   }

   private void createType(int token)
      throws ParsingException
   {
      currentSymbol.type = parseType(token);
   }

   private void createSymbol(int token, int kind)
      throws ParsingException
   {
      // make sure that we are syntactically correct
      parseIdentifier(token);

      // make sure no duplicates in this scope
      PascalSymbol s = symbolTable.getSymbol(ps.currentTokenValue, 
                                             kind, 
                                             false);

      if (s != null)
         throw new TypeException(s);
      
      // create a new symbol for this identifier
      currentSymbol = new PascalSymbol(ps.currentTokenValue, kind);

      // add to the end of the global symbol table
      symbolTable.add(currentSymbol);
   }
   
   /**
    * This function will parse ONE variable type declaration
    */
   private void parseVars(int token, boolean isSpecial)
      throws ParsingException
   {
      List newVariables = (List) new Vector();

      int currentToken = token; 

      if (currentToken == PascalLanguage.TOKEN_IDENT)
      {
         createSymbol(currentToken, PascalLanguage.KIND_VAR);
   
         // add newly created object to receive the final type
         newVariables.add(currentSymbol);
  
         // May be VAR parameter, if so mark as such
         currentSymbol.setSpecial(isSpecial);

         while (GetSym() == PascalLanguage.TOKEN_COMMA)
         {
            createSymbol(GetSym(), PascalLanguage.KIND_VAR);

            // add newly created object to receive the final type
            newVariables.add(currentSymbol);

            // May be VAR parameter, if so mark as such
            currentSymbol.setSpecial(isSpecial);
         }
         
         currentToken = ps.currentToken;   

         if (currentToken != PascalLanguage.TOKEN_COLON)
            throw new SyntaxException(currentToken, PascalLanguage.TOKEN_COLON);
   
         createType(GetSym());

         // assign that type to all of the declared variables
         ListIterator li = newVariables.listIterator();
         while (li != null && li.hasNext())
         {
            PascalSymbol newSym = (PascalSymbol)li.next();
            newSym.type = currentSymbol.type;
         }
      }
   }

   private void parseFormalParameters(int token)
      throws ParsingException
   {
      int currentToken = token;
      boolean isVarParameter = false;

      if (currentToken != PascalLanguage.TOKEN_LP)
         throw new SyntaxException(currentToken, 
                                   PascalLanguage.TOKEN_LP);

      while ((currentToken = GetSym()) == PascalLanguage.TOKEN_IDENT ||
             currentToken == PascalLanguage.TOKEN_VAR)
      {
         if (currentToken == PascalLanguage.TOKEN_VAR)
         {
            isVarParameter = true;
            currentToken = GetSym(); 
         }
         else
            isVarParameter = false;

         parseVars(currentToken, isVarParameter);

         if (GetSym() != PascalLanguage.TOKEN_SEMICOLON)
            putbackToken = true;
      }

      if (currentToken != PascalLanguage.TOKEN_RP)
         throw new SyntaxException(currentToken, 
                                   PascalLanguage.TOKEN_RP);

      currentToken = GetSym();
      
      if (currentToken == PascalLanguage.TOKEN_COLON)
      {
         currentToken = GetSym();

         if (currentToken != PascalLanguage.TOKEN_IDENT)
            throw new SyntaxException(currentToken, 
                                      PascalLanguage.TOKEN_IDENT);

         // FIXME: do right thing
         String returnType = ps.currentTokenValue;
      }
      else
         putbackToken = true;
   }

   private void parseProcedureDeclarations(int token)
      throws ParsingException
   {
      int currentToken = token;
      int formalParameterSize = 0;
      int symbolSize = 0;

      while (currentToken == PascalLanguage.TOKEN_PROCEDURE)
      {
         if (GetSym() != PascalLanguage.TOKEN_IDENT)
            throw new SyntaxException(ps.currentToken,
                                      PascalLanguage.TOKEN_IDENT);

         String procName = ps.currentTokenValue;
         //System.out.println(procName);
         /*
          * Fixup the procedure start point later.
          */
         PascalSymbol proc = new PascalSymbol(procName, 
                                              PascalLanguage.KIND_PROC, 
                                              0);

         proc.procSymbols.setParent(symbolTable);
         symbolTable.add(proc);

         PascalSymbolTable oldSymbolTable = symbolTable;
         symbolTable = proc.procSymbols;

         parseFormalParameters(GetSym());
         proc.formalParameters = symbolTable.size();
         formalParameterSize = symbolTable.getSymbolTableSize();

         if (GetSym() != PascalLanguage.TOKEN_SEMICOLON)
            throw new SyntaxException(ps.currentToken,
                                      PascalLanguage.TOKEN_SEMICOLON);

         parseDeclarations(GetSym());
      
         // We need to have the symbol table defined BEFORE we parse
         // the procedures

         // Local variable size
         proc.addressOffset = symbolTable.getSymbolTableSize() - 
                              formalParameterSize;
         symbolTable.generateAddresses(null, 
                     -symbolTable.getSymbolTableSize()+ta.SIZE_INTEGER);
         symbolTable.reverseAddresses();

         // Parse procedure declarations (optional)
         parseProcedureDeclarations(GetSym());

         if (GetSym() == PascalLanguage.TOKEN_BEGIN)
         {
            proc.address = ta.getInstructionCount();

            parseStatementSequence(GetSym());
            clearRegisters();

            ta.genOp(ta.INST_RET, 0, 0, ta.REG_RET);
         }

         symbolTable = oldSymbolTable;

         if (GetSym() != PascalLanguage.TOKEN_END)
            throw new SyntaxException(ps.currentToken,
                                      PascalLanguage.TOKEN_END);

         currentToken = GetSym(); 

         if (!ps.currentTokenValue.equals(procName))
         {
            throw new SyntaxException(ps.currentToken,
                                      PascalLanguage.TOKEN_END);
         }

         if (GetSym() != PascalLanguage.TOKEN_SEMICOLON)
            throw new SyntaxException(ps.currentToken,
                                      PascalLanguage.TOKEN_SEMICOLON);

         currentToken = GetSym(); 
      }

      // we finished, so rollback one
      putbackToken = true;
   }

   private void parseDeclarations(int token)
      throws ParsingException
   {
      int currentToken = ps.currentToken;

      while (currentToken == PascalLanguage.TOKEN_CONST ||
            currentToken == PascalLanguage.TOKEN_TYPE ||
            currentToken == PascalLanguage.TOKEN_VAR)
      {
         switch (currentToken)
         {
         case PascalLanguage.TOKEN_CONST:
            while (GetSym() == PascalLanguage.TOKEN_IDENT)
            {
               currentToken = ps.currentToken;

               createSymbol(currentToken, PascalLanguage.KIND_CONST);
         
               // Pascal-2001 Constants are always INTEGER
               currentSymbol.type = new PascalType(
                                       PascalLanguage.TYPE_INTEGER,
                                       TargetArchitecture.SIZE_INTEGER);

               PascalSymbol constantSymbol = currentSymbol;

               if (GetSym() != PascalLanguage.TOKEN_EQUALS)
                  throw new SyntaxException(ps.currentToken, PascalLanguage.TOKEN_EQUALS);
   
               parseExpression(GetSym());

               if (currentSymbol.kind != PascalLanguage.KIND_CONST && 
                   currentSymbol.kind != PascalLanguage.KIND_HUGECONST)
                  throw new TypeException(constantSymbol, currentSymbol);

               // House the real constant value here...
               constantSymbol.address = currentSymbol.address;

               // Some numbers are too big to fit in an operand.
               if (constantSymbol.address > ta.MAX_INT)
                  constantSymbol.kind = PascalLanguage.KIND_HUGECONST;
   
               if (GetSym() != PascalLanguage.TOKEN_SEMICOLON)
                  throw new SyntaxException(ps.currentToken, PascalLanguage.TOKEN_SEMICOLON);
   
            }
            break;
         case PascalLanguage.TOKEN_TYPE:
            while (GetSym() == PascalLanguage.TOKEN_IDENT)
            {
               currentToken = ps.currentToken;

               createSymbol(currentToken, PascalLanguage.KIND_TYPE);
               
               if (GetSym() != PascalLanguage.TOKEN_EQUALS)
                  throw new SyntaxException(ps.currentToken, PascalLanguage.TOKEN_EQUALS);

               createType(GetSym());         

               if (GetSym() != PascalLanguage.TOKEN_SEMICOLON)
                  throw new SyntaxException(ps.currentToken, PascalLanguage.TOKEN_SEMICOLON);

            }
            break;
         case PascalLanguage.TOKEN_VAR:
            while (GetSym() == PascalLanguage.TOKEN_IDENT)
            {
               currentToken = ps.currentToken;

               parseVars(currentToken, false);

               if (GetSym() != PascalLanguage.TOKEN_SEMICOLON)
                  throw new SyntaxException(ps.currentToken, PascalLanguage.TOKEN_SEMICOLON);
            }
            break;
         default:
            // Should never happen
            throw new SyntaxException(currentToken, PascalLanguage.TOKEN_CONST);
         }

         currentToken = ps.currentToken;
      }

      putbackToken = true;
   }  

   private void parseExpression(int token)
      throws ParsingException
   {
      parseSimpleExpression(token);

      // Save the symbol   
      PascalSymbol leftSym = currentSymbol;

      int compToken = GetSym();
   
      int index = ta.arrayIndexOf(ta.comparisonTokens, compToken);

      if (index != -1)
      {
         parseSimpleExpression(GetSym());   

         leftSym = optimizeSymbols(ta.INST_CMP, leftSym, currentSymbol);

         // we compared two constant expressions.
         // There are no branch instructions that work with a constant,
         // so we need to load the constant into a register and then 
         // execute the branch normally.
         if (leftSym.kind == PascalLanguage.KIND_CONST)
            loadSymbol(leftSym);

         if (conditionNotToken == PascalLanguage.TOKEN_NOT)
            ta.genOp(ta.comparisonInstructions[index], leftSym.registerNumber, 0, 0);
         else
            ta.genOp(ta.comparisonAntiInstructions[index], leftSym.registerNumber, 0, 0);

         // we are done with the symbol
         unloadSymbol(leftSym);
      }
      else
         putbackToken = true;

   }

   private void parseProcedure(int token)
      throws ParsingException
   {
      // Register window implementation
      List savedRegisters;

      //saveRegisters(registerTable);
      //savedRegisters = new Vector(registerTable);
      savedRegisters = new Vector();

      // If we are not a procedure, die.
      if (currentSymbol.kind != PascalLanguage.KIND_PROC)
         throw new TypeException(currentSymbol);

      // Get the procedure and all of its information
      PascalSymbol procedure = currentSymbol;
      PascalSymbolTable procParamsTable = procedure.getSymbols();
      PascalSymbolTable excludeRestoreParams = new PascalSymbolTable();
      ListIterator paramSyms = procParamsTable.listIterator();
      PascalSymbol currentParameter = null;
      List actualItems = new Vector();
      int actualParamCount = 0;

      // Parse and verify the procedure's actual parameters
      if (token == PascalLanguage.TOKEN_LP)
      {
         while (GetSym() != PascalLanguage.TOKEN_RP)
         {
            parseExpression(ps.currentToken);
            PascalItem actualItem = new PascalItem(currentSymbol, currentType);
            actualItems.add(actualItem);

            if (!paramSyms.hasNext())
               throw new TypeException(procedure, currentSymbol);

            currentParameter = (PascalSymbol)paramSyms.next();

            if (currentParameter.type.compareTo(actualItem.type) != 0)
               throw new TypeException(currentParameter, actualItem.symbol);

            // Before we push it, we might need to load it.
            // Consider parameters with expressions (x+4).
            // However, we need to exclude ARRAY and RECORD
            // as they are treated specially when pushing.
            if (actualItem.type.form != PascalLanguage.TYPE_ARRAY &&
                actualItem.type.form != PascalLanguage.TYPE_RECORD)
               loadSymbol(actualItem.symbol);

            pushSymbol(actualItem);

            actualParamCount++;
            if (GetSym() != PascalLanguage.TOKEN_COMMA)
               putbackToken = true;
         }
      }
      else
         putbackToken = true;

      // We should have no parameters left
      if (actualParamCount != procedure.formalParameters)
         throw new TypeException(procedure);

      clearRegisters();
      // Setup SP to point at end of local variables
      // Frame Pointer will point here in a second
      ta.genOp(ta.INST_SUBI, ta.REG_SP, ta.REG_SP, procedure.addressOffset);

      // Return address and old frame pointer go at top of
      // stack (lowest number) ABOVE new FP

      int staticScopes = symbolTable.getSymbolTableDepth();
      if (staticScopes <= 0)
         staticScopes = 0;
      else
      {
         int tempReg = allocateRegister();
         for (int i = 1; i <= staticScopes; i++)
         {
            ta.genOp(ta.INST_LDW, tempReg, ta.REG_FP, -(i*ta.SIZE_INTEGER));
            ta.genOp(ta.INST_PSH, tempReg,  ta.REG_SP, -ta.SIZE_INTEGER);
         }
         ta.deallocateRegister(tempReg);
      }

      ta.genOp(ta.INST_PSH, ta.REG_FP,  ta.REG_SP, -ta.SIZE_INTEGER);
      ta.genOp(ta.INST_PSH, ta.REG_RET, ta.REG_SP, -ta.SIZE_INTEGER);

      ta.genOp(ta.INST_ADDI, ta.REG_FP, ta.REG_SP, 
               (2 + staticScopes) * ta.SIZE_INTEGER);

      // Jump to the new location
      ta.genOp(ta.INST_BSR, 0, 0, procedure.address - ta.getInstructionCount() );

      // Restore FP and Ret register
      ta.genOp(ta.INST_POP, ta.REG_RET, ta.REG_SP, ta.SIZE_INTEGER);
      ta.genOp(ta.INST_POP, ta.REG_FP,  ta.REG_SP, ta.SIZE_INTEGER);

      for (int i = 1; i <= staticScopes; i++)
         ta.genOp(ta.INST_POP, 0,  ta.REG_SP, ta.SIZE_INTEGER);

      // Restore SP
      ta.genOp(ta.INST_ADDI, ta.REG_SP, ta.REG_SP, procedure.addressOffset);

      for (int i = actualItems.size() - 1; i >= 0; i--)
      {
         Object o = actualItems.get(i);
         if (o instanceof PascalItem)
         {
            PascalItem pi = (PascalItem)o;
            
            PascalSymbol formalSym = (PascalSymbol) procParamsTable.get(i);

            // Was it previously loaded into our registers?
            // If so, don't restore that register later on.
            if (formalSym.isSpecial() && registerTable.contains(pi.symbol))
               excludeRestoreParams.add(pi.symbol);

            popSymbol(pi, !formalSym.isSpecial());
         }
      }

      restoreRegisters(savedRegisters, excludeRestoreParams);
   }

   private void parseAssignment(int token)
      throws ParsingException
   {
      parseSymbol(token);

      PascalSymbol leftSym = currentSymbol;
      PascalType leftType = currentType;
  
      token = GetSym();

      // We may have stumbled upon a procedure.  Handle it.
      // WriteLn does not have any actual parameters.
      // FIXME: Is this a bug?
      if (token != PascalLanguage.TOKEN_ASSIGN)
      {
         parseProcedure(token);
         return;
      } 

      parseExpression(GetSym());

      if (leftType.compareTo(currentType) != 0)
         throw new TypeException(leftSym, currentType);

      copyData(leftSym, leftType, currentSymbol, currentType);
   }

   private void parseIfStatement(int token)
      throws ParsingException
   {
      // We have to ensure a solid state of registers for each
      // step.  There are ways to get around this, but it isn't
      // the worst thing in the world right now.
      clearRegisters();

      // There are two kinds of fixups: immediate and complete.
      // We either need to move on to the next check
      // or we have completed a statement sequence
      // and want to move on.
      // fixupAddress handles immediates,
      // and completeFixup handles completes.
      int fixupAddress = -1;

      // i dunno...
      int[] oldFixups = fixups;

      fixups = new int[FIXUP_LENGTH];
      //Arrays.fill(fixups,0);

      // in case anyone cares, indicate we are in a IF stmt.
      int oldConditionToken = conditionIfToken;
      conditionIfToken = PascalLanguage.TOKEN_IF;

      parseExpression(GetSym());

      // instruction we just did - i.e branch
      fixupAddress = ta.getInstructionCount() - 1;

      if (GetSym() != PascalLanguage.TOKEN_THEN)
         throw new SyntaxException(ps.currentToken, PascalLanguage.TOKEN_THEN);
     
      clearImmediateFixups();

      parseStatementSequence(GetSym());
      clearRegisters();

      ta.genOp(ta.INST_BEQ, 0, 0, 0);
      markCompleteFixup();

      int currentToken = GetSym();

      while (currentToken == PascalLanguage.TOKEN_ELSIF)
      {
         ta.fixupInstruction(fixupAddress, ta.getInstructionCount());

         parseExpression(GetSym());

         // instruction we just did - i.e branch
         fixupAddress = ta.getInstructionCount() - 1;

         if (GetSym() != PascalLanguage.TOKEN_THEN)
            throw new SyntaxException(ps.currentToken, PascalLanguage.TOKEN_THEN);
      
         clearImmediateFixups();
         parseStatementSequence(GetSym());
         clearRegisters();

         ta.genOp(ta.INST_BEQ, 0, 0, 0);
         markCompleteFixup();

         currentToken = GetSym();
      }

      // FIXME: If there is no ELSE, how about deleting the last
      // instruction?
      ta.fixupInstruction(fixupAddress, ta.getInstructionCount());
      //clearImmediateFixups();

      if (currentToken == PascalLanguage.TOKEN_ELSE)
      {
         parseStatementSequence(GetSym());

         // update it.
         currentToken = GetSym();
      }
      
      // If ELSE was called, this will do something,
      // otherwise we are back in steady state now.
      clearRegisters();

      if (currentToken == PascalLanguage.TOKEN_END)
      {
         clearMayFixups();
         clearCompleteFixups();
      }
      else
         throw new SyntaxException(ps.currentToken, PascalLanguage.TOKEN_END);

      conditionIfToken = oldConditionToken;

      fixups = oldFixups;
   }

   private void parseWhileStatement(int token)
      throws ParsingException
   {
      // FIXME: ouch, ouch, ouch.
      // We need to ensure a consistent state of the memory
      // because we may have to come back to the instruction
      // later.  At this point, I'll just clear the registers
      // out.
      clearRegisters();

      // There are two kinds of fixups: immediate and complete.
      // We either need to move on to the next check
      // or we have completed a statement sequence
      // and want to move on.
      // fixupAddress handles immediates,
      // and completeFixup handles completes.
      // i dunno...
      int[] oldFixups = fixups;

      fixups = new int[FIXUP_LENGTH];

      // in case anyone cares, indicate we are in a IF stmt.
      // FIXME: While == IF in this case...
      int oldConditionToken = conditionIfToken;
      conditionIfToken = PascalLanguage.TOKEN_IF;

      int loopAddress = ta.getInstructionCount();

      parseExpression(GetSym());

      // The last instruction should go to the end.
      markCompleteFixup();

      // Mark all fixups that would start the loop now
      clearImmediateFixups();

      if (GetSym() != PascalLanguage.TOKEN_DO)
         throw new SyntaxException(ps.currentToken, PascalLanguage.TOKEN_DO);
      
      parseStatementSequence(GetSym());

      if (GetSym() != PascalLanguage.TOKEN_END)
         throw new SyntaxException(ps.currentToken, PascalLanguage.TOKEN_END);

      // Before we go back to the beginning, we need to clear the
      // registers again.
      clearRegisters();

      // We need to go back to the beginning.
      ta.genOp(ta.INST_BEQ, 0, 0, loopAddress - ta.getInstructionCount());

      // clear all complete fixups to point to now.
      clearMayFixups();
      clearCompleteFixups();

      conditionIfToken = oldConditionToken;

      fixups = oldFixups;
   }

   private void parseStatement(int token)
      throws ParsingException
   {
      if (token == PascalLanguage.TOKEN_IDENT)
         parseAssignment(token);
      else if (token == PascalLanguage.TOKEN_WRITE)
         parseWrite(token);
      else if (token == PascalLanguage.TOKEN_WRITELN)
         parseWriteln(token);
      else if (token == PascalLanguage.TOKEN_IF)
         parseIfStatement(token);
      else if (token == PascalLanguage.TOKEN_WHILE)
         parseWhileStatement(token);
      else // We don't want to get involved
         putbackToken = true;

      // At the end of each statement, we can clear all unnamed
      // variables.
      clearUnnamedRegisters();
   }

   private void checkStatement(int token)
      throws ParsingException
   {
      try
      {
         parseStatement(token);
      }
      catch (TypeException te)
      {
         System.err.println("Error " + ++typeErrorCount + ":\n" );
         //System.err.println(te.toString());
         //System.err.println("\n");
         te.printStackTrace();

         //System.out.println("Error detected - see stderr");
         //te.printStackTrace();
         //System.out.println(ps.currentTokenValue);

         // This is absurd...
         // It also makes the debug output look funny.  Nice.
         // Oh, well.
         int currentToken = ps.currentToken;
         while (currentToken != PascalLanguage.TOKEN_SEMICOLON &&
               currentToken != PascalLanguage.TOKEN_END)
            currentToken = GetSym();

         putbackToken = true;
      }
   }

   private void parseStatementSequence(int token) 
      throws ParsingException
   {
      // always rollback once - always doing the GetSym makes 
      // it cleaner 
      putbackToken = true;

      do {
         checkStatement(GetSym());   
      }
      while (GetSym() == PascalLanguage.TOKEN_SEMICOLON);

      putbackToken = true;
   }
   
   private void parseProgram(int token) 
      throws ParsingException
   {
      parseIdentifier(GetSym());

      if (GetSym() != PascalLanguage.TOKEN_SEMICOLON)
         throw new SyntaxException(ps.currentToken, PascalLanguage.TOKEN_SEMICOLON);

      // We will branch unconditionally to the main statement sequence.
      ta.genOp(ta.INST_BEQ, 0, 0, 0);
      int fixupAddress = ta.getInstructionCount() - 1;

      emitPredefinedFunctions();

      parseDeclarations(GetSym());   

      // We need to have the symbol table defined BEFORE we parse
      // the procedures
      symbolTable.generateAddresses(ta);
      //symbolTable.printSymbolTable();
      //System.out.println(symbolTable.getSymbolTableDepth());

      // Parse procedure declarations (optional)
      parseProcedureDeclarations(GetSym());

      if (GetSym() != PascalLanguage.TOKEN_BEGIN)
         throw new SyntaxException(ps.currentToken, PascalLanguage.TOKEN_BEGIN);
  
      // The next instruction to generate will be the first instruction
      // of our main code.  So, branch to here. 
      ta.fixupInstruction(fixupAddress, ta.getInstructionCount());
      // setup the machine
      ta.genOp(ta.INST_ADD, ta.REG_FP, ta.REG_GLOBAL, 0);
      int stackOffset = symbolTable.getSymbolTableSize();

      if (stackOffset >= ta.SIZE_INTEGER)
         stackOffset -= ta.SIZE_INTEGER;

      ta.genOp(ta.INST_SUBI, ta.REG_SP, ta.REG_GLOBAL, stackOffset);

      //parse program
      parseStatementSequence(GetSym());   

      // FIXME: As the code execute different paths, the register
      // table changes.  However, our code is only done with one
      // pass.  We are certainly going to overrun the register
      // table with this algorithm.  Need something similar to
      // the PascalSymbolTable to handle register allocation.
      // Think If-Then-Else constructs...
      clearRegisters();
      
      if (GetSym() != PascalLanguage.TOKEN_END)
         throw new SyntaxException(ps.currentToken, PascalLanguage.TOKEN_END);

      // We have now completed this program.
      // FIXME: Could do a RET to R31, then do a RET 0 there.  Not now.
      ta.genOp(ta.INST_RET, 0, 0, 0);
   }

   /**
    * Parse/compile the source file.
    *
    * @param S Source file name
    */
   public void parseFile(String s)
   {
      StringBuffer sb = ps.readFile(s);
      ps.setInputStream(sb);   

      // Start State is a Program
      try
      {
         parseProgram(GetSym());
      }
      catch (ParsingException pe)
      {
         pe.printStackTrace();
      }
   }

   /**
    * Print out relevant information about the file we just
    * compiled.
    *
    * @see TargetArchitecture
    */
   public void writeFile()
   {
      ta.writeObjectFile();
   }

   public void compile()
   {
      parseFile(filename); 
      writeFile();
   }

   /**
    * Test driver entry point.
    *
    * @param args   Command-line parameters
    */
   public static void main(String[] args)
   {
      if (args.length == 0 || args[0] == null || args[0].equals(""))
      {
         System.out.println("java PascalParser <Input Filename>");
         return;
      }

      PascalParser pp = new PascalParser(args[0]);
      pp.compile();
   }
}
