import java.util.Vector;
import java.util.ListIterator;

/**
 * Table representing the allocated registers.
 */
public class PascalRegisterTable extends Vector
{
   TargetArchitecture ta;

   public PascalRegisterTable(TargetArchitecture newTA)
   {
      super();

      ta = newTA;
   }

   public void saveRegisters()
   {
      ListIterator li = listIterator();
      while (li != null && li.hasNext())
      {
         Object o = li.next();
         if (o instanceof PascalSymbol)
         {
            PascalSymbol sym = (PascalSymbol)o;
            ta.genOp(ta.INST_PSH, sym.registerNumber, ta.REG_SP, -ta.SIZE_INTEGER);
         }
      }
   }

   /**
    * Restores registers from the register table EXCEPT for those
    * contained within the excludes list.
    *
    * @param registerList The register table.
    * @param excludes The Exclude list for restoring from the register table.
    */
   public void restoreRegisters(PascalSymbolTable excludes)
   {
      for (int i = 1; i <= size(); i++)
      {
         Object o = get(size() - i);
         if (o instanceof PascalSymbol)
         {
            PascalSymbol sym = (PascalSymbol)o;
            if (!excludes.contains(sym))
               ta.genOp(ta.INST_POP, sym.registerNumber, ta.REG_SP, ta.SIZE_INTEGER);
            else
               ta.genOp(ta.INST_POP, 0, ta.REG_SP, ta.SIZE_INTEGER);
         }
      }
   }

   public void restoreRegisters()
   {
      restoreRegisters(new PascalSymbolTable());
   }

   /**
    * Goes through and invalidates ALL registers.
    *
    * NOTE: They are NOT stored back into memory.
    */
   public void invalidateRegisters()
   {
      // Clear out everything else now.
      ListIterator li = registerTable.listIterator();
      while (li != null && li.hasNext())
      {
         Object o = li.next();
         if (o instanceof PascalSymbol)
         {
            PascalSymbol sym = (PascalSymbol)o;
            ta.deallocateRegister(sym.registerNumber);
            sym.registerNumber = ta.REG_INVALID;
            li.remove();
         }
      }
   }

   /**
    * Allocate register for a particular symbol.
    *
    * May force deallocation of other registers if full.
    *
    * This function should not be directly unless you know
    * what you are doing.
    *
    * @param s PascalSymbol that needs to be loaded
    */
   private void allocateRegister(PascalSymbol s)
   {
      if (s.registerNumber != TargetArchitecture.REG_INVALID)
         return;

      s.registerNumber = ta.allocateRegister();

      // Registers are full - Force run reaper...
      if (s.registerNumber == TargetArchitecture.REG_INVALID)
      {
         clearRegisters();

         s.registerNumber = ta.allocateRegister();

         // we are in bad, bad, bad shape...
         if (s.registerNumber == TargetArchitecture.REG_INVALID)
            return;
      }

      registerTable.add(s);
   }

   /**
    * Swap the two symbols in the register table.
    *
    * @param l Left symbol
    * @param r Right symbol
    */
   private void swapRegisters(PascalSymbol l, PascalSymbol r)
   {
      // nothing to do...
      if (l.registerNumber == TargetArchitecture.REG_INVALID &&
         r.registerNumber == TargetArchitecture.REG_INVALID)
         return;

      int lReg = l.registerNumber;
      int rReg = r.registerNumber;
   
      // if both are valid, they both need to stay in table   
      if (lReg == TargetArchitecture.REG_INVALID)
      {
         registerTable.remove(r);
         registerTable.add(l);
      }
      else if (rReg == TargetArchitecture.REG_INVALID)
      {
         registerTable.remove(l);
         registerTable.add(r);
      }

      l.registerNumber = rReg;
      r.registerNumber = lReg;
   }

   /**
    * This function takes the register assigned to the right
    * symbol and gives it to the left one.
    *
    * In cases where the right symbol is tied to a named value,
    * it must be written as it no longer controls the value.
    *
    * Optimizations could be made here to delay this until
    * the left symbol is modified by another operation, then
    * you can proceed to store the old value into the right
    * symbol, and then overwrite the new one.
    *
    * @param l Left symbol
    * @param r Right symbol
    */
   private void switchRegister(PascalSymbol l, PascalSymbol r)
   {
      // nothing to do...
      if (r.registerNumber == TargetArchitecture.REG_INVALID)
      {
         // we have an assignment, but it can be referred to as 
         // a constant, assign a register to the left side,
         // and do an ADDI with that value.
         if (r.kind == PascalLanguage.KIND_CONST)
         {
            if (l.registerNumber != TargetArchitecture.REG_INVALID)
               ta.deallocateRegister(l.registerNumber);
         
            allocateRegister(l);
            ta.genOp(ta.INST_ADDI, l.registerNumber, 0, r.address);
         }

         return;
      }

      int lReg = l.registerNumber;
      int rReg = r.registerNumber;
   
      // if the left one isn't in, add it.
      // we might be overriding an old value
      if (lReg == TargetArchitecture.REG_INVALID)
         registerTable.add(l);
      else
      {
         ta.deallocateRegister(l.registerNumber);
      }

      // the unload function will also deallocate the register
      // we just want to remove the named symbols's link in the 
      // register table and store it's value.
      if (r.isNamed() == true && r.kind == PascalLanguage.KIND_VAR)
      {
         storeSymbol(r);

         r.registerNumber = ta.REG_INVALID;
      }

      registerTable.remove(r);

      l.registerNumber = rReg;
   }

   /**
    * Loads generic symbol into memory.
    *
    * @param s Symbol to be loaded
    */
   private void loadSymbol(PascalSymbol s)
   {
      // already loaded
      if (s.registerNumber != ta.REG_INVALID)
         return;

      switch (s.kind)
      {
      case PascalLanguage.KIND_VAR:
         allocateRegister(s);

         // array offset support
         int loadAddress = s.address;

         // our addressing scheme is based on negative values.
         switch (s.type.form)
         {
         case PascalLanguage.TYPE_ARRAY:
         case PascalLanguage.TYPE_RECORD:
            if (s.type.link == null)
            {
               loadAddress -= s.addressOffset;
               ta.genOp(ta.INST_LDW, s.registerNumber, 
                        ta.getVarReg(), loadAddress);
            }
            else
               ta.genOp(ta.INST_LDX, s.registerNumber, 
                        ta.getVarReg(), s.type.link.registerNumber);
            break;
         default:
            ta.genOp(ta.INST_LDW, s.registerNumber, 
                     ta.getVarReg(), loadAddress);
            break;
         }
         break;
      case PascalLanguage.KIND_CONST:
         allocateRegister(s);
         ta.genOp(ta.INST_ADDI, s.registerNumber, 0, s.address);
         break;
      default:
         throw new UnsupportedOperationException(s.toString());
      }
   }

   /**
    * Stores symbol contents back in memory
    */
   private void storeSymbol(PascalSymbol s)
   {
      if (s.registerNumber == ta.REG_INVALID)
         return;

      int storeAddress = s.address;

      switch (s.type.form)
      {
      case PascalLanguage.TYPE_ARRAY:
      case PascalLanguage.TYPE_RECORD:
         // s.type.link == null if constant index
         // r.type.link != null if variable-based index
         if (s.type.link == null)
         {
            storeAddress -= s.addressOffset;
            ta.genOp(ta.INST_STW, s.registerNumber, ta.getVarReg(), storeAddress);
         }
         else
         {
            ta.genOp(ta.INST_STX, s.registerNumber, ta.getVarReg(), s.type.link.registerNumber );

            // We are done with our index register now.
            // The reaper should remove it.
            s.type.link = null;
         }
         break;
      default:
         ta.genOp(ta.INST_STW, s.registerNumber, ta.getVarReg(), storeAddress);
      }
   }

   /**
    * Unloads a generic symbol from memory.
    *
    * Named symbols need to be stored when this occurs.
    * Unnamed symbols can be silently discarded.
    *
    * NOTE: IF YOU CALL THIS FUNCTION, YOU MUST
    * DELETE IT FROM registerTable!
    *
    * @param s Symbol to be unloaded
    * @see unloadSymbol
    */
   private void discardSymbol(PascalSymbol s)
   {
      // If we loaded it somehow, then we will unload it
      // without question
      if (s.registerNumber == ta.REG_INVALID)
         return;

      if (s.isNamed() == true && s.kind == PascalLanguage.KIND_VAR)
         storeSymbol(s);

      ta.deallocateRegister(s.registerNumber);

      s.registerNumber = ta.REG_INVALID;
   }

   /**
    * Removes and frees up a register.
    *
    * @see discardSymbol
    */
   private void unloadSymbol(PascalSymbol s)
   {
      discardSymbol(s);
      registerTable.remove(s);
   }

   /**
    * Goes through all registers and unloads all of them.
    */
   private void clearRegisters()
   {
      if (registerTable != null)
      {  
         // due to KIND_LINK types, we MUST
         // unload named registers first. 
         ListIterator li = registerTable.listIterator();
         while (li != null && li.hasNext())
         {
            Object o = li.next();
            if (o instanceof PascalSymbol)
            {
               PascalSymbol sym = (PascalSymbol)o;
               if (sym.isNamed() == true)
               {
                  discardSymbol(sym);
                  li.remove();
               }
            }
         }

         // Clear out everything else now.
         li = registerTable.listIterator();
         while (li != null && li.hasNext())
         {
            Object o = li.next();
            if (o instanceof PascalSymbol)
            {
               PascalSymbol sym = (PascalSymbol)o;
               discardSymbol(sym);
               li.remove();
            }
         }
      }
   }

   /**
    * Goes through all registers and clears all unnamed symbols.
    *
    * Good for relieving register pressure from temporary symbols.
    */
   private void clearUnnamedRegisters()
   {
      if (registerTable != null)
      {   
         ListIterator li = registerTable.listIterator();
         while (li != null && li.hasNext())
         {
            Object o = li.next();
            if (o instanceof PascalSymbol)
            {
               PascalSymbol sym = (PascalSymbol)o;
               if ((!sym.isNamed() && !sym.isLinked()) || sym.isSpecial())
               {
                  discardSymbol(sym);
                  li.remove();
               }
            }
         }
      }
   }

   /**
    * This function will handle loading of symbols and
       * optimizing constants.
    *
    * FIXME: It will not handle unloading of symbols - yet.
    *
    * @param opCode   The operation wish to be performed (no immediates!)
    * @param l    Left symbol
    * @param r    Right symbol
    *
    * @return The returned symbol.
    */
   private PascalSymbol optimizeSymbols(int opCode, 
                    PascalSymbol l, PascalSymbol r)
   {
      PascalSymbol s = new PascalSymbol(PascalLanguage.KIND_VAR);   

      // yes, the memory addresses may be equivalent   
      if (l != null && l != r && 
          l.kind == PascalLanguage.KIND_CONST &&
          r.kind == PascalLanguage.KIND_CONST &&
          l.type.compareTo(r.type) == 0)
      {
         // we need to now override the kind
         s.kind = PascalLanguage.KIND_CONST;

         // Assume the types are identical
         s.type = l.type;

         // These are our constant optimizers.
         switch (opCode)
         {
         case TargetArchitecture.INST_ADD:
            s.address = l.address + r.address;
            break;
         case TargetArchitecture.INST_SUB:
            s.address = l.address - r.address;
            break;
         case TargetArchitecture.INST_MUL:
            s.address = l.address * r.address;
            break;
         case TargetArchitecture.INST_DIV:
            s.address = l.address / r.address;
            break;
         case TargetArchitecture.INST_MOD:
            s.address = l.address % r.address;
            break;
         case TargetArchitecture.INST_CMP:
            s.address = (l.address == r.address) ? 1 : 0;
            break;
         case TargetArchitecture.INST_OR:
            // FIXME: Verify these
            s.address = l.address | r.address;
            break;
         case TargetArchitecture.INST_AND:
            // FIXME: Verify these
            s.address = l.address & r.address;
            break;
         case TargetArchitecture.INST_BIC:
            // FIXME: Verify these
            s.address = l.address & r.address;
            break;
         case TargetArchitecture.INST_XOR:
            // FIXME: Verify these
            s.address = l.address ^ r.address;
            break;
         case TargetArchitecture.INST_LSH:
            // FIXME: Verify these
            s.address = l.address << r.address;
            break;
         case TargetArchitecture.INST_RSH:
            // FIXME: Verify these
            s.address = l.address >> r.address;
            break;
         case TargetArchitecture.INST_CHK:
            // FIXME: Verify these
            s.address = (l.address == r.address) ? 1 : 0;
            break;
         default:
            throw new UnsupportedOperationException();
         }
            
      }
      else
      {
         // In the special case of a unary symbol (+<ident>),
         // we fake the left symbol by temporarily creating
         // a new symbol and assigning it register 0.
         // it will fall out of scope almost immediately.
         if (l == null)
         {
            l = new PascalSymbol();

            // yes.
            l.registerNumber = 0;

            // for below.
            l.type = r.type;
         }
         // Otherwise, insure that the symbol is loaded
         else if (l.registerNumber == TargetArchitecture.REG_INVALID) 
            loadSymbol(l);
      
         allocateRegister(s);   
         s.type = l.type;

         if (r.kind == PascalLanguage.KIND_CONST)
            ta.genOp(opCode + 16, s.registerNumber, l.registerNumber, r.address);
         else
         {
            if (r.registerNumber == TargetArchitecture.REG_INVALID) 
               loadSymbol(r);
      
            ta.genOp(opCode, s.registerNumber, l.registerNumber, r.registerNumber); 
         }

         // To be faithful to the doctrines given by
         // Assignment 7, we should always unload
         // the symbols after we are done.
         // Instead, we delay this until we are either
         // out of registers or we have exited the
         // statement sequence.
         // Per discussion with Dr. Franz, this is fine.
         //unloadSymbol(l);   
         //unloadSymbol(r);   
      }

      return s;   
   }

   public String toString()
   {
      StringBuffer sb = new StringBuffer(100);

      ListIterator li = listIterator();
      while (li != null && li.hasNext())
      {
         Object o = li.next();
         if (o instanceof PascalSymbol)
         {
            PascalSymbol sym = (PascalSymbol)o;
            sb.append(sym.registerNumber + ":\n" + sym);
         }
      }

      return sb.toString();
   }

}
