/**
 * This class defines the symbols.
 * 
 * Based on Dr. Franz's structure as described in lecture.
 */
public class PascalSymbol implements Comparable
{
   /**
    * The name associated with the symbol
    *
    * Will be blank for unnamed types
    */
   String identifier;

   /**
    * See PascalLanguage.KIND_
    */
   int kind;

   /**
    * What type are we?
    */
   PascalType type;
      
   /**
    * Where am I stored in memory?
    *
    * If type.form == TYPE_ARRAY, consult addressOffset 
    * as well.
    */
   int address;

   /**
    * Temporary offset to indicate where the exact location
    * of the memory address that I want to access is.
    *
    * (only valid if type.form == TYPE_ARRAY)
    */
   int addressOffset;
   
   /**
    * If allocated, which register was I last in?
    */   
   int registerNumber;

   /**
    * Special mode for VAR and other variables that should
    * not be stored in a register.
    */
   boolean specialMode;

   /**
    * Symbol table for procedures.
    *
    * (only valid for KIND_PROC)
    */
   PascalSymbolTable procSymbols;

   /**
    * Number of formal parameters we have.
    */
   int formalParameters;

   public PascalSymbol(String newIdentifier, int newKind, 
                     PascalType newType, int newAddress)
   {
      constructObject(newIdentifier, newKind, newType, newAddress);
   }

   public PascalSymbol(String newIdentifier, int newKind, 
                     PascalType newType)
   {
      constructObject(newIdentifier, newKind, newType, 0);
   }

   public PascalSymbol(int newKind, PascalType newType, int newAddress)
   {
      constructObject("", newKind, newType, newAddress);
   }

   public PascalSymbol(int newKind, PascalType newType)
   {
      constructObject("", newKind, newType, 0);
   }

   public PascalSymbol(String newIdentifier, int newKind)
   {
      constructObject(newIdentifier, newKind, null, 0);
   }

   public PascalSymbol(String newIdentifier, int newKind, 
                     int address)
   {
      constructObject(newIdentifier, newKind, null, address);
   }

   public PascalSymbol(int newKind)
   {
      constructObject("", newKind, null, 0);
   }

   public PascalSymbol(String newIdentifier)
   {
      constructObject(newIdentifier, PascalLanguage.KIND_INVALID, null, 0);
   }

   public PascalSymbol()
   {
      constructObject("", PascalLanguage.KIND_INVALID, null, 0);
   }

   private void constructObject(String newIdentifier, int newKind, 
                  PascalType newType, int newAddress)
   {
      identifier = newIdentifier;
      kind = newKind;
      type = newType;
      address = newAddress;
      specialMode = false;

      // Defaults to Invalid Register
      registerNumber = TargetArchitecture.REG_INVALID;

      // Defaults to base...
      addressOffset = 0;

      if (kind == PascalLanguage.KIND_PROC)
         procSymbols = new PascalSymbolTable();
      else
         procSymbols = null;
   }
   
   /**
    * @return True if named, False if unnamed
    */   
   public boolean isNamed()
   {
      return (identifier != null && !identifier.equals(""));
   }

   /**
    * @return True if linked to another symbol, False if not
    */   
   public boolean isLinked()
   {
      return (kind == PascalLanguage.KIND_LINK);
   }

   public boolean isSpecial()
   {
      return (specialMode);
   }

   public void setSpecial(boolean b)
   {
      specialMode = b;
   }

   public void setSpecial()
   {
      specialMode = true; 
   }

   /**
    * Adds a symbol to a procedure's symbol table.
    *
    * @param ps Symbol to add
    * @exception UnsupportedOperationException Thrown if used with non
    * Procedure
    */
   public void addSymbol(PascalSymbol ps)
   {
      if (kind != PascalLanguage.KIND_PROC)
         throw new UnsupportedOperationException();

      procSymbols.add(ps);  
   }

   /**
    * Retrieves the procedure's symbol table.
    *
    * @return PascalSymbolTable The symbol table
    * @exception UnsupportedOperationException Thrown if used with non
    * Procedure
    */
   public PascalSymbolTable getSymbols()
   {
      if (kind != PascalLanguage.KIND_PROC)
         throw new UnsupportedOperationException();
   
      return procSymbols;
   }

   public String toString()
   {
      StringBuffer sb = new StringBuffer(100);

      sb.append("Name:\t" + identifier);
      sb.append("\tKind:\t" + kind);
      sb.append("\tType:\t" + type); 

      if (kind == PascalLanguage.KIND_CONST ||
          kind == PascalLanguage.KIND_HUGECONST ||
          kind == PascalLanguage.KIND_VAR)
         sb.append("\tAddr:\t" + address);

      if (type != null && 
          type.form == PascalLanguage.TYPE_ARRAY)
         sb.append("\tOff:\t" + addressOffset);

      if (registerNumber != TargetArchitecture.REG_INVALID)
         sb.append("\tReg:\t" + registerNumber);

      return sb.toString();
   }

   public int compareTo(Object o)
   {
      if (o instanceof PascalSymbol)
      {
         PascalSymbol s = (PascalSymbol)o;
         int returnValue = identifier.compareTo(s.identifier);

         if (returnValue == 0)
            returnValue = type.compareTo(s.type);

         return returnValue;
      }
      else
         throw new ClassCastException();
   }
}
